;; Generated package description from zig-mode.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "zig-mode" "0.0.8.0.20260717.49" "A major mode for the Zig programming language" '((emacs "26.1") (reformatter "0.6")) :commit "62bfbaced0222e2bfbc086fa8556adf6b3298476" :authors '(("Andrea Orru" . "andreaorru1991@gmail.com") ("Andrew Kelley" . "superjoe30@gmail.com")) :maintainer '("Jen-Chieh" . "jcs090218@gmail.com") :keywords '("zig" "languages") :url "https://codeberg.org/ziglang/zig-mode")
